/*
 * $Id$
 *
 * Replace
 *
 *   To: <sip:user@email.ch>
 *
 * header in a SIP message with
 *
 *   To: <sip:32978@test.ethz.ch>
 *
 * where 32978 is the telephoneNumber attribute in LDAP for user with
 * primary email user@email.ch
 *
 *
 * Copyright (C) 2004 Swiss Federal Institute of Technology (ETH Zurich)
 * Copyright (C) 2004 Marcel Baur <baur@ethworld.ethz.ch>
 *
 * This module is released under the GNU General Public License.
 * Visit http://www.ethworld.ethz.ch for updated and further information.
 *
 */

/*
 * This file is an addition to ser, a free SIP server.
 *
 * Copyright (C) 2001-2003 Fhg Fokus
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * History:
 * --------
 * 2004-12-01 Initial version (Marcel Baur)
 */




#include "../../sr_module.h"
#include "../../action.h"
#include "../../db/db.h"
#include <stdio.h>
#include <stdlib.h>
#include <ldap.h>

MODULE_VERSION static int ldap_lookup_f (struct sip_msg *, char *, char *);
static int ldap_add_f (struct sip_msg *, char *, char *);

static int mod_init (void);

/* module parameters */

static char *LDAP_Server = "ldap.my.org";
static char *LDAP_Base = "o=my,c=org";
static char *LDAP_Attribute = "telephoneNumber";
static char *APPEND_Domain = "my.org";
static char *TABLE_SETUP = "ethworld_setup";
static char *FIELD_EMAIL = "email";
static char *FIELD_INETPHONENO = "inetPhoneNo";
static int LDAP_Port = 389;

static cmd_export_t cmds[] = {
  {"ldap_lookup", ldap_lookup_f, 0, 0, REQUEST_ROUTE},
  {"ldap_add", ldap_add_f, 0, 0, REQUEST_ROUTE},
  {0, 0, 0, 0, 0}
};

/* ------------------------------------------------------------------ */

static param_export_t params[] = {
  {"ldap_server", STR_PARAM, &LDAP_Server},
  {"ldap_port", INT_PARAM, &LDAP_Port},
  {"ldap_base", STR_PARAM, &LDAP_Base},
  {"ldap_attribute", STR_PARAM, &LDAP_Attribute},
  {"append_domain", STR_PARAM, &APPEND_Domain},
  {"table_setup", STR_PARAM, &TABLE_SETUP},
  {"field_email", STR_PARAM, &FIELD_EMAIL},
  {"field_inetPhoneNo", STR_PARAM, &FIELD_INETPHONENO},
  {0, 0, 0}
};

/* ------------------------------------------------------------------ */

struct module_exports exports = {
  "ldap",
  cmds,
  params,

  mod_init,			/* module initialization function */
  0,				/* response function */
  0,				/* destroy function */
  0,				/* oncancel function */
  0				/* per-child init function */
};

/* ------------------------------------------------------------------ */

#define RESULT_YES 1
#define RESULT_DROP_PACKET 0
#define RESULT_NO -1

static db_con_t *db_con;

static char *db_url = DEFAULT_DB_URL;
static db_func_t * db_f;


/* ------------------------------------------------------------------ */

static int
mod_init (void)
{
  int res;
  LOG (L_ERR, "LDAP Module\n"
       "Copyright (C) 2004 Swiss Federal Institute of Technology (ETH Zurich)\n"
       "Copyright (C) 2004 Marcel Baur <baur@ethworld.ethz.ch>\n");

  res = bind_dbmod (db_url, db_f);
  if (res < 0)
    {
      LOG (L_ERR, "Unable to bind to db module (err=%i)\n", res);
      return RESULT_NO;
    }
  db_con = db_f->init (db_url);
  if (db_con == NULL)
    {
      LOG (L_ERR, "ldap: Error in db_init");
      return RESULT_NO;
    }
  return 0;
}


/* ------------------------------------------------------------------ */
/* given a header field 'To: <sip:user@host>', 
 * lookup email user@host in LDAP and
 * return sip url 'sip:36297@my.org'
 */

static char *
get_filter (str to)
{
  char tmp[255];
  int start, stop;
  int i;

  start = 0;
  i = 0;

  if (strncmp (to.s, "sip:", 4) == 0)
    {
      start = 0;
    }
  else
    {

      while ((to.s[i] != '<') && (i < to.len))
	{
	  start++;
	  i++;
	}
      start++;
    }

  while ((to.s[i] != '>') && (i < to.len))
    {
      i++;
    }
  stop = i;

  if ((to.s[start] == 's') && (to.s[start + 1] == 'i')
      && (to.s[start + 2] == 'p') && (to.s[start + 3] = ':'))
    {
      start += 4;
    }

  if (start > stop)
    {
      LOG (L_ERR, "ldap: invalid To URI");
      return "";
    }

  strcpy (tmp, "(mail=");
  strncat (tmp, &to.s[start], stop - start);
  strcat (tmp, ")\0");
  return strdup (tmp);
}


/* ------------------------------------------------------------------ */
/* alter msg so uri is called */

static int
set_uri (const char *uri, struct sip_msg *msg)
{
  struct action act;
  int ret;

  memset (&act, 0, sizeof (act));
  act.type = SET_URI_T;
  act.p1_type = STRING_ST;
  act.p1.string = strdup (uri);
  ret = do_action (&act, msg);

  if (ret < 0)
    {
      LOG (L_ERR, "ERROR: unable to set uri (error=%i)\n", ret);
      ret = RESULT_NO;
    }
  else
    {
      ret = RESULT_YES;
    }
  return ret;
}


/* ------------------------------------------------------------------ */
/* append uri to dest set */

static int
append_uri (const char *uri, struct sip_msg *msg)
{
  struct action act;
  int ret;

  memset (&act, 0, sizeof (act));
  act.type = APPEND_BRANCH_T;
  act.p1_type = STRING_ST;
  act.p1.string = strdup (uri);
  ret = do_action (&act, msg);

  if (ret < 0)
    {
      LOG (L_ERR, "ERROR: unable to set uri (error=%i)\n", ret);
      ret = RESULT_NO;
    }
  else
    {
      ret = RESULT_YES;
    }
  return ret;
}


/* ------------------------------------------------------------------ */
/* search a number in ldap and if found, replace to: */

static char *
ldap_search_number (char *filter)
{
  LDAP *ld;
  LDAPMessage *msg, *entry;
  char *dn;
  int res, count;
  char *attrs[2];
  BerElement *ber;
  char *attr;
  char **vals;
  char result[255];

  result[0] = '\0';
  ld = ldap_init (LDAP_Server, LDAP_Port);

  dn = NULL;
  attrs[0] = LDAP_Attribute;
  attrs[1] = NULL;

  if (ld == NULL)
    {
      LOG (L_ERR, "ldap: Unable to contact server '%s:%i'\n", LDAP_Server,
	   LDAP_Port);
    }

  res =
    ldap_search_s (ld, LDAP_Base, LDAP_SCOPE_SUBTREE, filter, attrs, 0, &msg);
  if (res != LDAP_SUCCESS)
    {
      LOG (L_ERR, "ldap: Error %i in search\n", res);
      return "";
    }

  count = ldap_count_entries (ld, msg);

  if (count < 1)
    {
      return "";
    }
  entry = ldap_first_entry (ld, msg);
  while (entry)
    {
      dn = ldap_get_dn (ld, entry);
      if (dn != NULL)
	{
	  ldap_memfree (dn);

	  attr = ldap_first_attribute (ld, entry, &ber);
	  while (attr != NULL)
	    {
	      vals = ldap_get_values (ld, entry, attr);
	      LOG (L_ERR, "%s: %s\n", attr, vals[0]);
	      strcpy (result, "sip:");
	      strcat (result, vals[0]);
	      strcat (result, "@");
	      strcat (result, APPEND_Domain);
	      ldap_value_free (vals);

	      attr = ldap_next_attribute (ld, entry, ber);
	    }
	  entry = ldap_next_entry (ld, msg);
	}
    }

  ldap_msgfree (msg);
  free (filter);
  LOG (L_ERR, "ldap_search_number: '%s'", result);
  return strdup (result);
}


/* ------------------------------------------------------------------ */

static int
ldap_lookup_f (struct sip_msg *mmm, char *m1, char *m2)
{
  struct to_body * receiver;
  char *filter, *result;
  int retval;

  receiver = get_to(mmm);
  retval = RESULT_NO;
  filter = get_filter (receiver->uri);
  result = ldap_search_number (filter);
  if ((result != NULL) && (result != "") && (strlen (result) > 0))
    {
      set_uri (result, mmm);
      retval = RESULT_YES;
    }
  else
    {
      /* nothing found */
      retval = RESULT_NO;
    }
  LOG (L_ERR, "ldap_lookup: %i", retval);
  return retval;
}

/* ------------------------------------------------------------------------ */

static char *
scan_header (const str to)
{
  char tmp[255];
  int start, stop;
  int i;

  start = 0;
  i = 0;

  if (strncmp (to.s, "sip:", 4) == 0)
    {
      start = 0;
    }
  else
    {
      while ((to.s[i] != '<') && (i < to.len))
	{
	  start++;
	  i++;
	}
      start++;
    }

  while ((to.s[i] != '>') && (to.s[i] != '@') && (i < to.len))
    {
      i++;
    }
  stop = i;

  if ((to.s[start] == 's') && (to.s[start + 1] == 'i')
      && (to.s[start + 2] == 'p') && (to.s[start + 3] = ':'))
    {
      start += 4;
    }

  strncpy (tmp, &to.s[start], stop - start);
  tmp[stop - start] = '\0';

  LOG (L_ERR, "scan_header is '%s'", tmp);
  return strdup (tmp);
}

/* ------------------------------------------------------------------------- */

static char *
get_email_by_inet (char *inet)
{
  /*we registered only 1 param, so we ignore str2 */
  int result, len;
  char email[255];
  db_key_t _k[2];
  db_val_t _v[2];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz;

  db_val_t *values;
  db_row_t *rows;
  int i;

  /* SELECT */
  _c[0] = FIELD_EMAIL;
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, TABLE_SETUP);

  /* WHERE */
  _k[0] = FIELD_INETPHONENO;
  VAL_TYPE (_v) = DB_STR;
  VAL_NULL (_v) = 0;
  VAL_STR (_v).s = inet;
  VAL_STR (_v).len = strlen (inet);
  n = 1;

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);
  if (result < 0)
    {
      LOG (L_ERR, "get_email_by_inet: error %i querying db\n", result);
      return NULL;
    }

  anz = RES_ROW_N (res);

  if (anz == 0)
    {
      /* no entry found */
      db_f->free_result (db_con, res);
      LOG (L_ERR, "No email entry for inet='%s'", inet);
      return NULL;
    }
  else
    {
      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);

      for (i = 0; i < anz; i++)
	{
	  len = VAL_STR (values).len;
	  if (len > sizeof (email))
	    {
	      len = sizeof (email);
	    }
	  strncpy (email, VAL_STR (values).s, len);
	}
    }

  db_f->free_result (db_con, res);
  return strdup (email);

}

/* ------------------------------------------------------------------ */

static int
ldap_add_f (struct sip_msg *mmm, char *m1, char *m2)
{
  struct to_body * receiver;
  char *inet, *result, *email;
  char filter[255];
  int retval;

  retval = RESULT_NO;
  receiver = get_to(mmm);
  inet = scan_header (receiver->uri);
  LOG (L_ERR, "inet ist '%s'", inet);

  email = get_email_by_inet (inet);
  if ((email == "") || (email == NULL))
    {
      return RESULT_NO;
    }
  LOG (L_ERR, "email is '%s'", email);

  strcpy (filter, "(mail=");
  strncat (filter, email, 220);
  strcat (filter, ")");

  LOG (L_ERR, "filter is '%s'", filter);

  result = ldap_search_number (filter);

  if ((result != "") && (strlen (result) > 0))
    {
      LOG (L_ERR, "ldap search result is '%s'", result);
      append_uri (result, mmm);
      retval = RESULT_YES;
    }
  else
    {
      /* nothing found */
      retval = RESULT_NO;
    }
  LOG (L_ERR, "ldap_add: %i", retval);
  return retval;
}


/* ------------------------------------------------------------------ */
